"""Data models for Coinbase data fetcher."""

from datetime import datetime, timedelta
from enum import StrEnum
from typing import Literal, Optional

import pandas as pd
from pandas import Timestamp
from pydantic import BaseModel, Field, field_validator

from coinbase_data_fetcher.progress import ProgressBar, NullProgressBar


def yesterday_ts() -> pd.Timestamp:
    """Get yesterday's timestamp."""
    return pd.Timestamp(datetime.now().date() - timedelta(days=1))


class Coins(StrEnum):
    BITCOIN = "bitcoin"
    ETHEREUM = "ethereum"
    SOLANA = "solana"
    LITECOIN = "litecoin"
    DOGECOIN = "dogecoin"
    WIF = "dogwifhat"


class CoinInfo(BaseModel):
    coin: Coins
    symbol: str
    start_date: Timestamp
    logo_url: str = ""
    
    model_config = {"arbitrary_types_allowed": True}
    
    def __init__(self, **data):
        super().__init__(**data)
        if self.logo_url == "":
            self.logo_url = f"https://cryptologos.cc/logos/thumbs/{self.coin}.png"


COIN_INFO = {
    Coins.BITCOIN: CoinInfo(coin=Coins.BITCOIN, symbol="BTC-USD", 
                            start_date=Timestamp("2015-07-20")),
    Coins.ETHEREUM: CoinInfo(coin=Coins.ETHEREUM, symbol="ETH-USD", 
                            start_date=Timestamp("2016-07-21")),
    Coins.SOLANA: CoinInfo(coin=Coins.SOLANA, symbol="SOL-USD", 
                            start_date=Timestamp("2021-05-24")),
    Coins.LITECOIN: CoinInfo(coin=Coins.LITECOIN, symbol="LTC-USD", 
                            start_date=Timestamp("2017-05-03")),
    Coins.DOGECOIN: CoinInfo(coin=Coins.DOGECOIN, symbol="DOGE-USD", 
                            start_date=Timestamp("2021-06-03")),
    Coins.WIF: CoinInfo(coin=Coins.WIF, symbol="WIF-USD", 
                            start_date=Timestamp("2024-11-13")),
}


class CoinDataModel(BaseModel):
    model_config = {"arbitrary_types_allowed": True}
    
    coin: Coins = Field(
        default=Coins.BITCOIN,
        title="Select Coin",
        description="Choose the cryptocurrency to analyze",
    )
    
    data_granularity: int = Field(
        default=3600,
        title="Data granularity",
        description="E.g. 5min for 5 minutes candles",
        json_schema_extra={
            "choices": {
                60: "1 min.",
                300: "5 min.", 
                900: "15 min.",
                3600: "1 hour"
            }
        }
    )
    
    start_date: pd.Timestamp = Field(
        default_factory=lambda: yesterday_ts() - pd.DateOffset(months=3),
        title="Start Date",
        description="Beginning of simulation"
    )
    
    end_date: pd.Timestamp = Field(
        default_factory=lambda: yesterday_ts(),
        title="End Date",
        description="End of simulation"
    )
    
    price_interpolation: Literal["Hi-Lo", "mean"] = Field(
        default="Hi-Lo",
        title="Price interpolation",
        description="""Hi-Lo: Use the high as the start price of a bearish candle 
        and the low in the middle between the following candle's start.
        For bullish candles vice-verse.
        
        Mean: Use the mean of the high and low of the candle as the 
        price for the entire candle period.""",
    )

    @field_validator('start_date', 'end_date', mode='before')
    @classmethod
    def parse_timestamp(cls, value):
        if isinstance(value, str):
            return pd.Timestamp(value).tz_localize(None)
        return value
    
    @classmethod
    def get_choices(cls, field_name: str) -> list:
        """Get choices for a field if available."""
        field = cls.model_fields.get(field_name)
        if field and hasattr(field, 'json_schema_extra'):
            extra = field.json_schema_extra
            if isinstance(extra, dict) and 'choices' in extra:
                return list(extra['choices'].keys())
        return []


class CoinData:
    
    def __init__(self, model: CoinDataModel):
        self.model = model
        
    def fetch_prices(self, progress_bar: Optional[ProgressBar] = None):
        from coinbase_data_fetcher.fetcher import fetch_prices
        
        if progress_bar is None:
            progress_bar = NullProgressBar()
            
        return fetch_prices(
            coin=self.model.coin,
            start_time=self.model.start_date,
            end_time=self.model.end_date,
            granularity=self.model.data_granularity,
            use_candle_hi_lo=self.model.price_interpolation == "Hi-Lo",
            progress_bar=progress_bar
        )


#!/usr/bin/env python3
"""Pre-fetch cryptocurrency data to warm the cache."""

import os
from datetime import datetime, timedelta
from typing import Optional

import pandas as pd

from coinbase_data_fetcher.config import config
from coinbase_data_fetcher.fetcher import fetch_prices
from coinbase_data_fetcher.models import COIN_INFO, CoinDataModel
from coinbase_data_fetcher.progress import TqdmProgressBar, NullProgressBar


def fetch_data_for_coin(coin, granularity, save_csv: bool = True, progress_bar_desc: Optional[str] = None, start_date: Optional[str] = None, end_date: Optional[str] = None, interpolate_price: bool = True):
    """Fetch data for a specific coin and granularity."""
    try:
        from coinbase_data_fetcher.progress import TqdmProgressBar
        progress_bar = TqdmProgressBar(
            total=100, 
            desc=progress_bar_desc or f"{coin.upper()}-{int(granularity/60)}m"
        )
    except ImportError:
        progress_bar = NullProgressBar()
    
    # Use provided dates or defaults
    start_time = pd.Timestamp(start_date) if start_date else COIN_INFO[coin].start_date
    yesterday = pd.Timestamp.now().date() - pd.Timedelta(days=1)
    
    if end_date:
        end_time = pd.Timestamp(end_date)
        # Ensure end date is not later than yesterday
        if end_time.date() > yesterday:
            print(f"Warning: End date {end_date} is in the future. Using yesterday ({yesterday}) instead.")
            end_time = yesterday
    else:
        end_time = yesterday
    
    df = fetch_prices(
        coin,
        start_time=start_time,
        end_time=end_time,
        granularity=granularity,
        progress_bar=progress_bar,
        leave_pure=not interpolate_price,
        use_candle_hi_lo=interpolate_price
    )
    
    if save_csv:
        # Write to CSV
        start_date = df.index[0].date().strftime('%Y-%m-%d')
        end_date = df.index[-1].date().strftime('%Y-%m-%d')
        
        # Create cache folder if not exists
        cache_path = config.cache_path
        if not os.path.exists(cache_path):
            os.makedirs(cache_path)
        
        csv_path = f'{cache_path}/{coin}_{granularity}_{start_date}_{end_date}.csv'
        df.to_csv(csv_path)
        print(f"Saved: {csv_path}")
    
    return df


def prefetch_all_data():
    """Pre-fetch all coin data for all granularities."""
    coins = CoinDataModel.get_choices("coin")
    granularities = CoinDataModel.get_choices("data_granularity")
    
    print(f"Pre-fetching data for {len(coins)} coins with {len(granularities)} granularities...")
    print(f"Cache directory: {config.cache_path}")
    
    for coin in coins:
        for granularity in granularities:
            try:
                fetch_data_for_coin(coin, granularity)
            except Exception as e:
                print(f"Error fetching {coin} at {granularity}s: {e}")
                continue
    
    print("Pre-fetching completed!")


def main():
    """CLI entry point for prefetching data."""
    import argparse
    
    parser = argparse.ArgumentParser(description="Pre-fetch cryptocurrency data")
    parser.add_argument('--coin', help="Specific coin to fetch (e.g., bitcoin)")
    parser.add_argument('--granularity', type=int, help="Specific granularity in seconds (e.g., 3600)")
    parser.add_argument('--start-date', help="Start date for fetching (e.g., 2023-01-01)")
    parser.add_argument('--end-date', help="End date for fetching (e.g., 2023-12-31)")
    parser.add_argument('--no-interpolate-price', action='store_true', help="Don't interpolate prices using candlestick hi/lo data")
    parser.add_argument('--no-csv', action='store_true', help="Don't save CSV files")
    parser.add_argument('--cache-path', help="Override cache directory")
    
    args = parser.parse_args()
    
    if args.cache_path:
        config.cache_path = args.cache_path
    
    if args.coin and args.granularity:
        # Fetch specific coin and granularity
        date_info = ""
        if args.start_date or args.end_date:
            date_info = f" from {args.start_date or 'start'} to {args.end_date or 'yesterday'}"
        print(f"Fetching {args.coin} data at {args.granularity}s granularity{date_info}...")
        fetch_data_for_coin(args.coin, args.granularity, save_csv=not args.no_csv, 
                           start_date=args.start_date, end_date=args.end_date,
                           interpolate_price=not args.no_interpolate_price)
    elif args.coin:
        # Fetch all granularities for specific coin
        granularities = CoinDataModel.get_choices("data_granularity")
        date_info = ""
        if args.start_date or args.end_date:
            date_info = f" from {args.start_date or 'start'} to {args.end_date or 'yesterday'}"
        print(f"Fetching all granularities for {args.coin}{date_info}...")
        for granularity in granularities:
            try:
                fetch_data_for_coin(args.coin, granularity, save_csv=not args.no_csv,
                                   start_date=args.start_date, end_date=args.end_date,
                                   interpolate_price=not args.no_interpolate_price)
            except Exception as e:
                print(f"Error fetching {args.coin} at {granularity}s: {e}")
                continue
    else:
        # Fetch all coins and granularities
        if args.start_date or args.end_date:
            print("Warning: --start-date and --end-date are ignored when fetching all data")
        prefetch_all_data()


if __name__ == "__main__":
    main()


"""Tests for models module."""

import pandas as pd
import pytest
from coinbase_data_fetcher.models import CoinData, CoinDataModel, CoinInfo, Coins, COIN_INFO


class TestCoins:
    def test_coin_enum_values(self):
        assert Coins.BITCOIN == "bitcoin"
        assert Coins.ETHEREUM == "ethereum"
        assert Coins.SOLANA == "solana"


class TestCoinInfo:
    def test_coin_info_creation(self):
        info = CoinInfo(
            coin=Coins.BITCOIN,
            symbol="BTC-USD",
            start_date=pd.Timestamp("2023-01-01")
        )
        assert info.coin == Coins.BITCOIN
        assert info.symbol == "BTC-USD"
        assert info.logo_url == "https://cryptologos.cc/logos/thumbs/bitcoin.png"
    
    def test_coin_info_with_custom_logo(self):
        info = CoinInfo(
            coin=Coins.ETHEREUM,
            symbol="ETH-USD",
            start_date=pd.Timestamp("2023-01-01"),
            logo_url="https://example.com/logo.png"
        )
        assert info.logo_url == "https://example.com/logo.png"


class TestCoinDataModel:
    def test_default_values(self):
        model = CoinDataModel()
        assert model.coin == Coins.BITCOIN
        assert model.data_granularity == 3600
        assert model.price_interpolation == "Hi-Lo"
        
    def test_parse_timestamp(self):
        model = CoinDataModel(
            start_date="2023-01-01",
            end_date="2023-12-31"
        )
        assert isinstance(model.start_date, pd.Timestamp)
        assert isinstance(model.end_date, pd.Timestamp)
        assert model.start_date.tz is None
        assert model.end_date.tz is None
    
    def test_get_choices(self):
        choices = CoinDataModel.get_choices("data_granularity")
        assert 60 in choices
        assert 300 in choices
        assert 3600 in choices


class TestCoinData:
    def test_coin_data_creation(self):
        model = CoinDataModel(coin=Coins.ETHEREUM)
        coin_data = CoinData(model)
        assert coin_data.model.coin == Coins.ETHEREUM